#ifndef _DATA_ACQUISITION_H_
#define _DATA_ACQUISITION_H_
 
/**************************************************************************
* Copyright (c) 2007
* Computer Aided Interventions and Medical Robotics (CAIMR) group,
* Imaging Science and Information Systems (ISIS) research center,
* Georgetown University.
*
* All rights reserved.
*
* Permission to use, copy, modify, distribute, and sell this software
* and its documentation for any purpose is hereby granted without fee,
* provided that:
* (1) Redistributions of source code must retain the above copyright notice,
*     this list of conditions and the following disclaimer.
* (2) Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
* (3) The author's name may not be used to endorse or promote products derived
*     from this software without specific prior written permission.
* (4) Modified source versions must be plainly marked as such, and must not be
*     misrepresented as being the original software.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ``AS IS'' AND ANY EXPRESS
* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
* OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

#include "TrackerCommunicationException.h"
#include "DataAcquisitionSettings.h"
#include <vector>
#include <igstkSerialCommunication.h>
#include <igstkTracker.h>
#include "IGSTKErrorObserver.h"

class DataAcquisitionSettingsFileReaderException;


/**
 * The exception type thrown by the DataAcquisition methods.
 *
 * @author Ziv Yaniv
 */
class DataAcquisitionException : public std::exception
{
public:
  /**
   * Construct an exception with a specifc message.
   */
  DataAcquisitionException(const std::string &errorMessage) {this->errorMessage = errorMessage;}

  /**
   * Virtual destructor.
   */
  virtual ~DataAcquisitionException() throw() {}

  /**
   * Get the error message.
   */
   virtual const char* what() const throw() {return errorMessage.c_str();}
private:
  std::string errorMessage;
};


/**
 * This is the class that actually runs the data acquisition. It inherits from itk::Object so that I
 * can use the Observer/Observable pattern implemented there. This way the data acquisition is seperated
 * from the UI. The UI can initiate data acquisition or stop it by directly invoking DataAcquisition methods, 
 * and it receives updates from the DataAcquisition by observing the following events: 
 * DataAcquisition::DataAcquistionStart(), DataAcquisition::DataAcquistionEnd(), DataAcquisition::DataAcquired(). 
 *
 *
 * The data is written to file as "TimeStamp x  y  z  qx  qy  qz qw" where [qx,qy,qz,qw] is a unit quaternion.
 * The time stamp is used for temporal synchronization between multiple files. This is required because only valid 
 * data is written to file. That is, the N samples in each file do not necessarily correspond temporaly. The time stamp
 * is in milliseconds.
 *
 * @author Ziv Yaniv
 */
class DataAcquisition : public itk::Object
{
public:
  /**
   * Create an instance of the DataAcquisition class.
   */
  DataAcquisition(DataAcquisitionSettings &settings);

  /**
   * Destroy the DataAcquisition object. This method replaces the standard destructor due to the use
   * of reference counting by the itk::Object (see documentation of itk::LightObject::Delete()).
   */
   virtual void Delete();

  /**
   * Open serial communication with the tracker according to the given communication settings, load all the tool
   * SROM files and start tracking.
   * @throw TrackerCommunicationException If there is a problem with the communication throw an exception.
   * @throw DataAcquisitionException If the method is called before the configuration file is loaded throw an exception.
   */
  void StartTracking() throw (TrackerCommunicationException);	         

  /**
   * Get the tracking data for the tools write it to file and notify all observers of the following events
   * DataAcquistionStart(), DataAcquisitionEnd(), DataAcquired(). File
   * names are port_x.i.txt where i is a running index and x is the tool port
   * number on the control box. The running index is incremented after every
   * invocation of AcquireData().
   *
   * @throw DataAcquistionException If there is a problem with the tracker update or the StartTracking method was not successfuly invoked
   *                                throw an exception.
   */
  void AcquireData() throw (DataAcquisitionException);

  /**
   * Same functionality as AcquireData() only this method overwrites the previous
   * data files.
   */
  void ReacquireData() throw (DataAcquisitionException);

  /**
   * Stop the data acquisition.
   * Note that this does not stop the tracking, we simply don't query the tracker and log
   * the data.
   */
  void StopAcquisition();

  /**
   * The event generated when data acquisition starts.
   */
  itkEventMacro(DataAcquisitionStartEvent, itk::StartEvent);   

  /**
   * The event generated when data acquisition ends.
   */
  itkEventMacro(DataAcquisitionEndEvent, itk::EndEvent);   

  /**
   * The event generated for every data acquisition iteration.
   */
  class  DataAcquiredEvent : public itk::IterationEvent {
  public:
    typedef DataAcquiredEvent Self;
    typedef itk::IterationEvent Superclass;
    DataAcquiredEvent() : percentage(0) {};
    DataAcquiredEvent(const std::string &information, double percentage) {this->information = information; this->percentage = percentage;}
    virtual ~DataAcquiredEvent() {} 
    virtual const char * GetEventName() const {return "DataAcquisition::DataAcquiredEvent"; } 
    virtual bool CheckEvent(const ::itk::EventObject* e) const {return dynamic_cast<const Self*>(e);}
    virtual ::itk::EventObject* MakeObject() const {return new Self;}
    DataAcquiredEvent(const Self& s) : itk::IterationEvent(s){this->percentage = s.percentage; this->information = s.information;}
    /**
     * Get a string describing the acquired data (port, channel and the translation and quaternion info).
     */
    std::string GetInformation() const {return this->information;}
    /**
     * How much of the total data has been acquired in [0,1].
     */
    double GetPercentage() const {return this->percentage;}
    virtual void Print(std::ostream& os) const {os<<this->percentage<<" percent done: "<<this->information<<"\n";}
   private:
     void operator=(const Self&);
     double percentage;
     std::string information;
  };


private:
           //do not allow destruction through the standard destructor, the user should call the Delete() method.
           //this is a constraint imposed by the itk::Object's use of refernce counting
  ~DataAcquisition() {}

        //tracker data
	igstk::Tracker::Pointer tracker;
	igstk::SerialCommunication::Pointer serialCommunication;
  static const igstk::Tracker::TimePeriodType DEFAULT_VALIDITY_TIME;
  igstk::Tracker::TimePeriodType validityTime; //a transformation is valid for a period of "validityTime" milliseconds

        //these are the tools whose tracking information we need to record
  std::vector<NDIToolSettings *> trackerTools; 

  unsigned int currentFileNumber; //running index used in the file names 
  int sampleNumber; //how many samples do we acquire in the AcquireData() and ReacquireData() methods
  bool stopAcquisition; 
                             //an array containing the names of the output files and the streams
  std::pair<std::string, std::ofstream> *outputFiles;
                            //an array containing the number of samples acquired for each tool
  int *numAcquiredSamples;
                        //total number of samples we need to acquire, numberOfTools*sampleNumber
  int sampleSize;
                      //this object will observe ALL igstk components for errors
  IGSTKErrorObserver::Pointer errorObserver;

  bool tracking;  //the Acquire(), ReAcquire() methods can only be called if tracking==true
};
#endif //_DATA_ACQUISITION_H_
